# cython: boundscheck=False
# cython: cdivision=True
# cython: wrapararound=False

cdef unsigned char NONE = 0
cdef unsigned char OFFSET = 1
cdef unsigned char REL_GAIN = 2
cdef unsigned char BPMASK = 4

from cython.parallel import prange


def correct_burst(
    unsigned short[:, :, :] image_data,
    unsigned char[:, :, :] gain_stage,
    unsigned char[:] cell_table,
    unsigned char flags,
    float[:, :, :, :] offset_map,
    float[:, :, :, :] relgain_map,
    unsigned[:, :, :, :] badpixel_mask,
    float badpixel_fill_value,
    float[:, :, :] output,
):
    cdef int input_cell, map_cell, x, y
    cdef unsigned char gain
    cdef float corrected
    for input_cell in prange(image_data.shape[0], nogil=True):
        map_cell = cell_table[input_cell]
        if map_cell >= offset_map.shape[0]:
            for y in range(image_data.shape[1]):
                for x in range(image_data.shape[2]):
                    output[input_cell, y, x] = <float>image_data[input_cell, y, x]
            continue
        for y in range(image_data.shape[1]):
            for x in range(image_data.shape[2]):
                corrected = image_data[input_cell, y, x]
                gain = gain_stage[input_cell, y, x]
                # legal values: 0, 1, or 3
                if gain == 2:
                    corrected = badpixel_fill_value
                else:
                    if gain == 3:
                        gain = 2

                    if (flags & BPMASK) and badpixel_mask[map_cell, y, x, gain] != 0:
                        corrected = badpixel_fill_value
                    else:
                        if (flags & OFFSET):
                            corrected = corrected - offset_map[map_cell, y, x, gain]
                        if (flags & REL_GAIN):
                            corrected = corrected / relgain_map[map_cell, y, x, gain]
                output[input_cell, y, x] = corrected


def correct_single(
    unsigned short[:, :, :] image_data,
    unsigned char[:, :, :] gain_stage,
    unsigned char flags,
    float[:, :, :] offset_map,
    float[:, :, :] relgain_map,
    unsigned[:, :, :] badpixel_mask,
    float badpixel_fill_value,
    float[:, :, :] output,
):
    # ignore "cell table", constant pretends cell 0
    cdef int x, y
    cdef unsigned char gain
    cdef float corrected
    for y in range(image_data.shape[1]):
        for x in range(image_data.shape[2]):
            corrected = image_data[0, y, x]
            gain = gain_stage[0, y, x]
            # legal values: 0, 1, or 3
            if gain == 2:
                corrected = badpixel_fill_value
            else:
                if gain == 3:
                    gain = 2

                if (flags & BPMASK) and badpixel_mask[y, x, gain] != 0:
                    corrected = badpixel_fill_value
                else:
                    if (flags & OFFSET):
                        corrected = corrected - offset_map[y, x, gain]
                    if (flags & REL_GAIN):
                        corrected = corrected / relgain_map[y, x, gain]
            output[0, y, x] = corrected
